@@ -71,6 +71,24 @@ REGEX_HEADING = re.compile(r"""
                                 # whitespaces.
     $                           # Match line end.""", re.VERBOSE)
 
+REGEX_HEADING__ALTERNATIVE_LEVEL_1 = re.compile(r"""
+    (.+?)   # CAPTURE GROUP (1) | Match between 1 and ∞ characters, as few
+            # times as possible.
+    \n      # Match a newline.
+    \s*     # Match between 1 and ∞ whitespaces.
+    ={2,}   # Match "=" between 2 and ∞ times.
+    $       # Match line end.
+""", re.VERBOSE | re.MULTILINE)
+
+REGEX_HEADING__ALTERNATIVE_LEVEL_2 = re.compile(r"""
+    (.+?)   # CAPTURE GROUP (1) | Match between 1 and ∞ characters, as few
+            # times as possible.
+    \n      # Match a newline.
+    \s*     # Match between 1 and ∞ whitespaces.
+    -{2,}   # Match "-" between 2 and ∞ times.
+    $       # Match line end.
+""", re.VERBOSE | re.MULTILINE)
+
 REGEX_HORIZONTAL_RULE = re.compile(r"""
     ^               # Match line start.
     \s*             # Match between 0 and ∞ whitespaces.
@@ -144,6 +162,56 @@ REGEX_LINK = re.compile(r"""
     \s*         # Match between 0 and ∞ whitespaces.
     \)          # Match ")" once.""", re.VERBOSE)
 
+REGEX_REFERENCE_DEFINITION = re.compile(r"""
+    ^           # Match line start.
+    (?<!\\)     # Ensure there's no escaping backslash.
+    \s*         # Match between 0 and ∞ whitespaces.
+    \[          # Match "[" once.
+    \s*         # Match between 0 and ∞ whitespaces.
+    (.+?)       # CAPTURE GROUP (1) | Match between 1 and ∞ characters, as few
+                # times as possible.
+    \s*         # Match between 0 and ∞ whitespaces.
+    \]          # Match "]" once.
+    \s*         # Match between 0 and ∞ whitespaces.
+    :           # Match ":" once.
+    \s*         # Match between 0 and ∞ whitespaces.
+    <?          # Match "<" either 0 or 1 times.
+    \s*         # Match between 0 and ∞ whitespaces.
+    (.+?)       # CAPTURE GROUP (2) | Match between 1 and ∞ characters, as few
+                # times as possible.
+    \s*         # Match between 0 and ∞ whitespaces.
+    >?          # Match ">" either 0 or 1 times.
+    \s*         # Match between 0 and ∞ whitespaces.
+    (?:         # Open non-capturing group.
+        [\"'(]  # Match '"', "'", or "(" once.
+        \s*     # Match between 0 and ∞ whitespaces.
+        (.+?)   # CAPTURE GROUP (3) | Match between 1 and ∞ characters, as few
+                # times as possible.
+        \s*     # Match between 0 and ∞ whitespaces.
+        [\"')]  # Match '"', "'", or ")" once.
+    )?          # Close non-capturing group and match it either 0 or 1 times.
+    \s*         # Match between 0 and ∞ whitespaces.
+    $           # Match line end.""", re.VERBOSE | re.MULTILINE)
+
+REGEX_REFERENCE_LINK = re.compile(r"""
+    \s*         # Match between 0 and ∞ whitespaces.
+    (?<!\\)     # Ensure there's no escaping backslash.
+    (?:         # Open non-capturing group.
+        \[      # Match "[" once.
+        \s*     # Match between 0 and ∞ whitespaces.
+        (.+?)   # CAPTURE GROUP (1) | Match between 1 and ∞ characters, as few
+                # times as possible.
+        \s*     # Match between 0 and ∞ whitespaces.
+        \]      # Match "]" once.
+    )?          # Close non-capturing group and match it either 0 or 1 times.
+    \s*         # Match between 0 and ∞ whitespaces.
+    \[          # Match "[" once.
+    \s*         # Match between 0 and ∞ whitespaces.
+    (.+?)       # CAPTURE GROUP (2) | Match between 1 and ∞ characters, as few
+                # times as possible.
+    \s*         # Match between 0 and ∞ whitespaces.
+    \]          # Match "]" once.""", re.VERBOSE)
+
 REGEX_ORDERED_LIST = re.compile(r"""
     (\s+)?  # CAPTURE GROUP (1) | Match between 1 and ∞ whitespaces, as many
             # times as possible, as either one or zero matches.
@@ -155,6 +223,7 @@ REGEX_ORDERED_LIST = re.compile(r"""
     \s*     # Match between 0 and ∞ whitespaces.""", re.VERBOSE)
 
 REGEX_QUICK_EMAIL = re.compile(r"""
+    (?<!\\)         # Ensure there's no escaping backslash.
     <               # Match "<" once.
     (               # CAPTURE GROUP (1) | Open capture group.
         [\w\d_.+-]+ # Match either a word character, a digit, "_", ".", "+", or
@@ -169,6 +238,7 @@ REGEX_QUICK_EMAIL = re.compile(r"""
     >               # Match ">" once.""", re.VERBOSE)
 
 REGEX_QUICK_LINK = re.compile(r"""
+    (?<!\\)             # Ensure there's no escaping backslash.
     <                   # Match "<" once.
     (                   # CAPTURE GROUP (1) | Open capture group.
         https?          # Match either "http" or "https".
@@ -469,51 +539,64 @@ def convert_nested_tag(line, cur_tag, open_tags):
     return new_line
 
 
-def add_inline_tags(line):
+def add_inline_tags(line, references):
     """
     Add inline tags, such as <em> and <strong>, to a line.
 
     Args:
         line (str): Line to add tags to.
+        references (List[Dict]): A list of dictionaries, each containing
+            information about a reference-style link definition.
 
     Returns:
         line (str): Converted line.
     """
     # Add emphasis.
-    # The order here is important, otherwise "**bold**" would be converted
-    # to "*<em>bold</em>*", instead of "<strong>bold</strong>".
+    # The order here is important, otherwise "**bold**" would be converted to
+    # "*<em>bold</em>*", instead of "<strong>bold</strong>".
     line = REGEX_BOLD.sub("<strong>\\1\\2</strong>", line)
     line = REGEX_ITALIC.sub("<em>\\1\\2</em>", line)
 
     # Add images and links.
     # The order here is important, otherwise images wouldn't work.
     if REGEX_IMAGE.search(line):
-        match = REGEX_LINK.search(line)
-        alt_text, url, title = match.groups()
-        line = REGEX_IMAGE.sub(
-            f'<img src="{url}" alt="{alt_text}"'
-            + (f' title="{title}"' if title else '')
-            + '>', line)
+        matches = REGEX_LINK.findall(line)
+        for match in matches:
+            alt_text, url, title = match
+            line = REGEX_IMAGE.sub(
+                f'<img src="{url}" alt="{alt_text}"'
+                + (f' title="{title}"' if title else '')
+                + '>', line, 1)
     if REGEX_LINK.search(line):
-        match = REGEX_LINK.search(line)
-        alt_text, url, title = match.groups()
-        line = REGEX_LINK.sub(
-            f'<a href="{url}"'
-            + (f' title="{title}"' if title else '')
-            + f'>{alt_text}</a>', line)
+        matches = REGEX_LINK.findall(line)
+        for match in matches:
+            text, url, title = match
+            line = REGEX_LINK.sub(
+                f'<a href="{url}"'
+                + (f' title="{title}"' if title else '')
+                + f'>{text}</a>', line, 1)
 
-    # Add quick email links.
-    if REGEX_QUICK_EMAIL.search(line):
-        match = REGEX_QUICK_EMAIL.search(line)
-        address = match.group(1)
-        line = REGEX_QUICK_EMAIL.sub(
-            f'<a href="mailto:{address}">{address}</a>', line)
+    # Add reference-style links.
+    if references:
+        matches = REGEX_REFERENCE_LINK.findall(line)
+        for match in matches:
+            text, label = match
+            for reference in references:
+                if label == reference["label"]:
+                    label, url, title = reference.values()
+                    line = REGEX_REFERENCE_LINK.sub(
+                        f'<a href="{url}"'
+                        + (f' title="{title}"' if title else '')
+                        + f'>{text if text else label}</a>', line, 1)
+                    break
 
     # Add quick links.
     if REGEX_QUICK_LINK.search(line):
-        match = REGEX_QUICK_LINK.search(line)
-        address = match.group(1)
-        line = REGEX_QUICK_LINK.sub(f'<a href="{address}">{address}</a>', line)
+        line = REGEX_QUICK_LINK.sub("<a href=\"\\1\">\\1</a>", line)
+
+    # Add quick links to email addresses.
+    if REGEX_QUICK_EMAIL.search(line):
+        line = REGEX_QUICK_EMAIL.sub("<a href=\"mailto:\\1\">\\1</a>", line)
     return line
 
 
@@ -527,9 +610,19 @@ def convert(string):
     Returns:
         new_string (str): HTML code.
     """
+    # Store reference-style link definitions.
+    keys = ("label", "url", "title")
+    references = [
+        dict(zip(keys, i)) for i in REGEX_REFERENCE_DEFINITION.findall(string)]
+    string = REGEX_REFERENCE_DEFINITION.sub("", string)
+
     if string.strip() == "":
         return ""
 
+    # Convert alternate-style headings to conventional style.
+    string = REGEX_HEADING__ALTERNATIVE_LEVEL_1.sub("# \\1", string)
+    string = REGEX_HEADING__ALTERNATIVE_LEVEL_2.sub("## \\1", string)
+
     open_tags = []
     new_string = ""
     add_line_break = False
@@ -604,12 +697,12 @@ def convert(string):
             if code_snippets:
                 line = ""
                 for block in code_snippets:
-                    left = add_inline_tags(block["left"])
+                    left = add_inline_tags(block["left"], references)
                     line += left + f"<code>{block['content']}</code>"
                     if block == code_snippets[-1]:
-                        line += add_inline_tags(block["right"])
+                        line += add_inline_tags(block["right"], references)
             else:
-                line = add_inline_tags(line)
+                line = add_inline_tags(line, references)
 
             # Check if line contains nested tags, if so, open tags.
             if any(tag["regex"].fullmatch(line) for tag in NESTED_TAGS):
